"""Custom client for interacting with Google's Photos API"""

from datetime import datetime
from enum import Enum, auto
from json import dumps
from os import getenv
from os.path import join, isfile

from requests import get

from wg_utilities.clients._generic import GoogleClient
from wg_utilities.functions import user_data_dir, force_mkdir

LOCAL_MEDIA_DIRECTORY = getenv(
    "LOCAL_MEDIA_DIRECTORY", user_data_dir(file_name="media_downloads")
)


class MediaType(Enum):
    """Enum for all potential media types"""

    IMAGE = auto()
    VIDEO = auto()


class Album:
    """Class for Google Photos albums and their metadata/content

    Args:
        json (dict): the JSON which is returned from the Google Photos API when
         describing the album
        google_client (GooglePhotosClient): an active Google client which can be used to
         list media items

    """

    def __init__(self, json, google_client=None):
        self.json = json
        self._media_items = None

        self.google_client = google_client

    @property
    def media_items(self):
        # noinspection GrazieInspection
        """Lists all media items in the album

        Returns:
            list: a list of MediaItem instances, representing the contents of the album
        """

        if not self._media_items:
            self._media_items = self.google_client.get_album_contents(self.id)

        return self._media_items

    @property
    def title(self):
        """
        Returns:
            str: the title of the album
        """
        return self.json.get("title")

    @property
    def id(self):
        """
        Returns:
            str: the ID of the album
        """
        return self.json.get("id")

    @property
    def media_items_count(self):
        """
        Returns:
            int: the number of media items within the album
        """
        return int(self.json.get("mediaItemsCount", "-1"))

    def __str__(self):
        return f"{self.title}: {self.id}"


class MediaItem:
    """Class for representing a MediaItem and its metadata/content

    Args:
        json (dict): the JSON returned from the Google Photos API, which describes
        this media item
    """

    def __init__(self, json):
        self.json = json

        try:
            self.creation_datetime = datetime.strptime(
                json.get("mediaMetadata", {}).get("creationTime"),
                "%Y-%m-%dT%H:%M:%S.%fZ",
            )
        except ValueError:
            self.creation_datetime = datetime.strptime(
                json.get("mediaMetadata", {}).get("creationTime"), "%Y-%m-%dT%H:%M:%SZ"
            )

        self.local_path = join(
            LOCAL_MEDIA_DIRECTORY,
            self.creation_datetime.strftime("%Y/%m/%d"),
            json["filename"],
        )

    def download(self, width_override=None, height_override=None, force_download=False):
        """Download the media item to local storage. The width/height overrides do
        not apply to videos

        Args:
            width_override (int): the width override to use when downloading the file
            height_override (int): the height override to use when downloading the file
            force_download (bool): flag for forcing a download, even if it exists
             locally already
        """
        if not self.stored_locally or force_download:
            width = width_override or self.width
            height = height_override or self.height

            # LOGGER.debug("Downloading %s (%sx%s)", self.filename, width, height)

            param_str = {
                MediaType.IMAGE: f"=w{width}-h{height}",
                MediaType.VIDEO: "=dv",
            }.get(self.media_type, "")

            with open(force_mkdir(self.local_path, path_is_file=True), "wb") as fout:
                fout.write(get(f"{self.json['baseUrl']}{param_str}").content)

    @property
    def bytes(self):
        """Opens the local copy of the file (downloading it first if necessary) and
        reads the binary content of it

        Returns:
            bytes: the binary content of the file
        """
        if not self.stored_locally:
            self.download()

        with open(self.local_path, "rb") as fin:
            bytes_content = fin.read()

        return bytes_content

    @property
    def stored_locally(self):
        """
        Returns:
            bool: flag for if the file exists locally
        """
        return isfile(self.local_path)

    @property
    def filename(self):
        """
        Returns:
            str: the media item's file name
        """
        return self.json.get("filename")

    @property
    def height(self):
        """
        Returns:
            int: the media item's height
        """
        return int(self.json.get("mediaMetadata", {}).get("height", "-1"))

    @property
    def width(self):
        """
        Returns:
            int: the media item's width
        """
        return int(self.json.get("mediaMetadata", {}).get("width", "-1"))

    @property
    def media_type(self):
        """Determines the media item's file type from the JSON

        Returns:
            MediaType: the media type (image, video, etc.) for this item
        """
        mime_type = self.json.get("mimeType", "")

        return {
            "image" in mime_type: MediaType.IMAGE,
            "video" in mime_type: MediaType.VIDEO,
        }.get(True)

    def __str__(self):
        return dumps(self.json, indent=4, default=str)


class GooglePhotosClient(GoogleClient):
    """Custom client for interacting with the Google Photos API

    See Also:
        GoogleClient: the base Google client, used for authentication and common
         functions
    """

    BASE_URL = "https://photoslibrary.googleapis.com/v1"

    def __init__(
        self,
        project,
        scopes=None,
        client_id_json_path=None,
        creds_cache_path=None,
        access_token_expiry_threshold=60,
        logger=None,
    ):
        super().__init__(
            project,
            scopes,
            client_id_json_path,
            creds_cache_path,
            access_token_expiry_threshold,
            logger,
        )

        self._albums = None

    def get_album_contents(self, album_id):
        # noinspection GrazieInspection
        """Gets the contents of a given album in Google Photos

        Args:
            album_id (str): the ID of the album which we want the contents of

        Returns:
            list: a list of MediaItem instances, representing each item in the album
        """

        return [
            MediaItem(item)
            for item in self._list_items(
                self.session.post,
                f"{self.BASE_URL}/mediaItems:search",
                "mediaItems",
                params={"albumId": album_id},
            )
        ]

    def get_album_from_name(self, album_name):
        """Gets an album definition from the Google API based on the album name

        Args:
            album_name (str): the name of the album to find

        Returns:
            Album: an Album instance, with all metadata etc.

        Raises:
            FileNotFoundError: if the client can't find an album with the correct name
        """

        self.logger.info("Getting metadata for album `%s`", album_name)
        for album in self.albums:
            if album.title == album_name:
                return album

        raise FileNotFoundError(f"Unable to find album with name {album_name}")

    @property
    def albums(self):
        """Lists all albums in the active Google account

        Returns:
            list: a list of Album instances
        """

        if not self._albums:
            self._albums = [
                Album(res, self)
                for res in self._list_items(
                    self.session.get,
                    f"{self.BASE_URL}/albums",
                    "albums",
                )
            ]

        return self._albums



"""Custom client for interacting with Google's Drive API"""

from copy import deepcopy

from wg_utilities.clients._generic import GoogleClient


class File:
    """A file object within Google Drive

    Args:
        id (str): the unique ID of this file
        name (str): the name of the file
        parents (list): a list of parent directory IDs (should always be of length 1)
        google_client (GoogleDriveClient): a Google client for use in other requests

    Attributes:
        file_id (str): the unique ID of this file
        name (str): the name of the file
        parent_id (str): the unique ID of the parent directory
        google_client (GoogleDriveClient): a Google client for use in other requests

    """

    DESCRIPTION_FIELDS = [
        "appProperties",
        "capabilities",
        "contentHints",
        "contentRestrictions",
        "copyRequiresWriterPermission",
        "createdTime",
        "description",
        "driveId",
        "explicitlyTrashed",
        "exportLinks",
        "fileExtension",
        "folderColorRgb",
        "fullFileExtension",
        "hasAugmentedPermissions",
        "hasThumbnail",
        "headRevisionId",
        "iconLink",
        "id",
        "imageMediaMetadata",
        "isAppAuthorized",
        "kind",
        "lastModifyingUser",
        "linkShareMetadata",
        "md5Checksum",
        "mimeType",
        "modifiedByMe",
        "modifiedByMeTime",
        "modifiedTime",
        "name",
        "originalFilename",
        "ownedByMe",
        "owners",
        "parents",
        "permissionIds",
        "permissions",
        "properties",
        "quotaBytesUsed",
        "resourceKey",
        "shared",
        "sharedWithMeTime",
        "sharingUser",
        "shortcutDetails",
        "size",
        "spaces",
        "starred",
        "teamDriveId",
        "thumbnailLink",
        "thumbnailVersion",
        "trashed",
        "trashedTime",
        "trashingUser",
        "version",
        "videoMediaMetadata",
        "viewedByMe",
        "viewedByMeTime",
        "viewersCanCopyContent",
        "webContentLink",
        "webViewLink",
        "writersCanShare",
    ]

    # noinspection PyShadowingBuiltins
    # pylint: disable=redefined-builtin
    def __init__(self, id, name, parents=None, google_client=None, **_):
        self.file_id = id
        self.name = name
        self.parent_id = parents.pop() if parents else None
        self.google_client = google_client

        self._description = None
        self._parent = None

    def describe(self, force_update=False):
        """Describe the file by requesting all available fields from the Drive API

        Args:
            force_update (bool): re-pull the description from Google Drive, even if we
             already have the description locally

        Returns:
            dict: the description JSON for this file
        """
        if force_update or self._description is None:
            self._description = self.google_client.session.get(
                f"{self.google_client.BASE_URL}/files/{self.file_id}",
                params={"fields": ", ".join(self.DESCRIPTION_FIELDS)},
            ).json()

        return self._description

    @property
    def parent(self):
        """
        Returns:
            Directory: the parent directory of this file
        """
        if not self._parent:
            self._parent = self.google_client.get_directory_by(
                "file_id", self.parent_id
            )

        return self._parent

    @property
    def path(self):
        """
        Returns:
            str: the path to this file in Google Drive
        """
        current_path = self.name

        parent_dir = self.parent
        while True:
            current_path = "/".join([parent_dir.name, current_path])

            if not (next_parent := parent_dir.parent):
                break

            parent_dir = next_parent

        return "/" + current_path

    def __gt__(self, other):
        return self.name.lower() > other.name.lower()

    def __lt__(self, other):
        return self.name.lower() < other.name.lower()

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


class Directory(File):
    """A Google Drive directory - basically a File with extended functionality

    Attributes:
        children (set): the directories contained within this directory
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.children = set()
        self._files = None

    @property
    def files(self):
        """
        Returns:
            list: the list of files contained within this directory
        """
        if not self._files:
            self._files = [
                File(**directory, google_client=self.google_client)
                for directory in self.google_client.get_items(
                    f"{self.google_client.BASE_URL}/files",
                    "files",
                    params=dict(
                        pageSize="1000",
                        q="mimeType != 'application/vnd.google-apps.folder' and"
                        f" '{self.file_id}' in parents",
                        fields="nextPageToken, files(id, name, parents)",
                    ),
                )
            ]

        return self._files

    @property
    def parent(self):
        """
        Returns:
            Directory: the parent directory of this directory
        """
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def tree(self):
        """A simple copy of the Linux `tree` command, this builds a directory tree in
        text form for quick visualisation

        Returns:
            str: the full directory tree in text form
        """

        def build_sub_tree(parent_dir, level, block_pipes_at_levels=None):
            """Builds a subtree of a given directory

            Args:
                parent_dir (Directory): the directory to create the subtree of
                level (int): the depth level of this directory
                block_pipes_at_levels (list): a list of levels to block further pipes at
            """

            nonlocal output

            # Creating a deep copy means that when we go back up from the recursion,
            # the previous iteration still has the correct levels in the list
            block_pipes_at_levels = deepcopy(block_pipes_at_levels) or []

            for i, child_dir in enumerate(sorted(parent_dir.children)):
                prefix = "\n"

                # build out the spaces and pipes on this line, in such a way to
                # maintain continuity from the previous line
                for j in range(level):
                    prefix += " " if j in block_pipes_at_levels else "│"
                    prefix += "    "

                # if this is the last child
                if i + 1 == len(parent_dir.children):
                    prefix += "└"
                    block_pipes_at_levels.append(level)
                else:
                    prefix += "├"

                prefix += "─── "
                output += prefix + child_dir.name

                build_sub_tree(child_dir, level + 1, block_pipes_at_levels)

        output = self.name
        build_sub_tree(self, 0)

        return output

    def __str__(self):
        return self.name


class GoogleDriveClient(GoogleClient):
    """Custom client specifically for Google's Drive API

    Args:
        project (str): the name of the project which this client is being used for
        scopes (list): a list of scopes the client can be given
        client_id_json_path (str): the path to the `client_id.json` file downloaded
         from Google's API Console
        creds_cache_path (str): file path for where to cache credentials
        access_token_expiry_threshold (int): the threshold for when the access token is
         considered expired
        logger (RootLogger): a logger to use throughout the client functions

    """

    BASE_URL = "https://www.googleapis.com/drive/v3"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._root_directory = None

        self._directories = None

    def _build_directory_structure(self):
        """Build the complete tree of directories, including parent-child relationships
        by listing all directories and then iterating through them to build the
        relationships
        """
        self._directories = [self.root_directory]

        # List every single directory
        for directory in self.get_items(
            f"{self.BASE_URL}/files",
            "files",
            params=dict(
                pageSize="1000",
                q="mimeType = 'application/vnd.google-apps.folder'",
                fields="nextPageToken, files(id, name, parents)",
            ),
        ):
            self._directories.append(Directory(**directory, google_client=self))

        # Iterate through the directories, adding their parents and children as
        # applicable
        for directory in self._directories:
            if parent_dir := self.get_directory_by("file_id", directory.parent_id):
                directory.parent = parent_dir
                parent_dir.children.add(directory)

    def get_directory_by(self, attribute, value):
        """Get a Directory instance by any attribute

        Args:
            attribute (str): the name of the attribute to search for
            value (Any): the target value of the attribute

        Returns:
            Directory: the directory being searched for, if it was found
        """
        for directory in self.directories:
            if getattr(directory, attribute) == value:
                return directory

        return None

    def get_file_from_id(self, file_id, params=None):
        """Find a file by its UUID

        Args:
            file_id (str): the unique ID of the file
            params (dict): ny params to pass in the request

        Returns:
            File: the target file, if found
        """

        if not self._directories:
            self._build_directory_structure()

        params = params or {}

        if "fields" not in params:
            params["fields"] = "id, name, parents"

        return File(
            **self.session.get(
                f"{self.BASE_URL}/files/{file_id}", params=params
            ).json(),
            google_client=self,
        )

    @property
    def shared_drives(self):
        """
        Returns:
            list: a list of Shared Drives the current user has access to
        """
        return self.get_items(f"{self.BASE_URL}/drives", "drives")

    @property
    def directories(self):
        """
        Returns:
            list: a list of all directories in the user's Drive
        """
        if not self._directories:
            self._build_directory_structure()

        return self._directories

    @property
    def root_directory(self):
        """
        Returns:
            Directory: the user's root directory/main Drive
        """
        if not self._root_directory:
            self._root_directory = Directory(
                **self.session.get(f"{self.BASE_URL}/files/root").json(),
                google_client=self,
            )
            if not self._directories:
                self._build_directory_structure()

        return self._root_directory



"""Useful functions for working with JSON/dictionaries"""


def set_nested_value(*, json_obj, keys, target_value, final_key=None):
    """Update a nested value in a dictionary

    Args:
        json_obj (dict): the JSON object to update
        keys (list): a list of keys used to traverse the dictionary
        target_value (Any): the value to set at the given location/path
        final_key (str): the final key, the value of which we're actually setting
    """

    final_key = final_key or keys.pop()

    if len(keys) > 0:
        set_nested_value(
            json_obj=json_obj.get(keys.pop(0), {}),
            keys=keys,
            target_value=target_value,
            final_key=final_key,
        )
    else:
        json_obj[final_key] = target_value



"""Generic Oauth client to allow for reusable authentication flows/checks etc."""
from copy import deepcopy
from datetime import datetime
from json import dump, load, dumps
from logging import getLogger, DEBUG
from os import remove
from random import choice
from string import ascii_letters
from threading import Thread
from time import sleep
from time import time
from webbrowser import open as open_browser

from flask import Flask, request

# noinspection PyPackageRequirements
from google.auth.transport.requests import AuthorizedSession

# noinspection PyPackageRequirements
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import Flow
from jwt import decode, DecodeError
from requests import get, post

from wg_utilities.functions import user_data_dir, force_mkdir
from wg_utilities.loggers import add_stream_handler


class TempAuthServer:
    """Temporary Flask server for auth flows, allowing the auth code to be retrieved
    without manual intervention

    Args:
        name (str): the name of the application package
        host (str): the hostname to listen on
        port (int): the port of the webserver
        debug (bool): if given, enable or disable debug mode
        auto_run (bool): automatically run the server on instantiation
    """

    def __init__(
        self,
        name,
        host="0.0.0.0",
        port=5001,
        debug=False,
        auto_run=True,
    ):
        self.host = host
        self.port = port
        self.debug = debug

        self.app = Flask(name)
        self.app_thread = Thread(
            target=lambda: self.app.run(
                host=self.host, port=self.port, debug=self.debug, use_reloader=False
            )
        )

        self._request_args = {}

        self.create_endpoints()

        if auto_run:
            self.run_server()

    def create_endpoints(self):
        """Create all necessary endpoints. This is a single method (rather than one
        method per endpoint) because of the need to use `self.app` as the decorator
        """

        @self.app.route("/get_auth_code", methods=["GET"])
        def get_auth_code():
            """Endpoint for getting auth code from third party callback

            Returns:
                dict: simple response dict
            """
            self._request_args[request.path] = request.args
            return {
                "status_code": 200,
            }

        @self.app.route("/kill", methods=["GET"])
        def kill():
            """Workaround endpoint for killing the server from within a script

            Returns:
                dict: simple response dict

            Raises:
                RuntimeError: if the server can't be killed due to the runtime
                 environment
            """
            if (func := request.environ.get("werkzeug.server.shutdown")) is None:
                raise RuntimeError("Not running with the Werkzeug Server")
            func()
            return {
                "status_code": 200,
            }

    def wait_for_request(self, endpoint, max_wait=300, kill_on_request=False):
        """Wait for a request to come into the server for when it is needed in a
         synchronous flow

        Args:
            endpoint (str): the endpoint path to wait for a request to
            max_wait (int): how many seconds to wait before timing out
            kill_on_request (bool): kill/stop the server when a request comes through

        Returns:
            dict: the args which were sent with the request

        Raises:
            TimeoutError: if no request is received within the timeout limit
        """
        self._request_args[endpoint] = None

        start_time = datetime.now()
        while (
            time_elapsed := (datetime.now() - start_time).seconds
        ) < max_wait and not self._request_args.get(endpoint):
            sleep(1)

        if time_elapsed > max_wait:
            raise TimeoutError(
                f"No request received to {endpoint} within {max_wait} seconds"
            )

        if kill_on_request:
            self.stop_server()

        return self._request_args[endpoint]

    def run_server(self):
        """Run the local server"""
        self.app_thread.start()

    def stop_server(self):
        """Stops the local server by hitting its `/kill` endpoint

        See Also:
            self.create_endpoints: `/kill` endpoint
        """
        # noinspection HttpUrlsUsage
        get(f"http://{self.host}:{self.port}/kill")


class OauthClient:
    """Custom client for interacting with Oauth APIs, including all necessary/basic
    authentication functionality"""

    ACCESS_TOKEN_ENDPOINT = None
    BASE_URL = None
    CREDS_FILE_PATH = None

    def __init__(
        self,
        *,
        client_id,
        client_secret,
        redirect_uri="http://0.0.0.0:5001/get_auth_code",
        access_token_expiry_threshold=60,
        log_requests=False,
        creds_cache_path=None,
        logger=None,
    ):

        if not all([self.ACCESS_TOKEN_ENDPOINT, self.BASE_URL, self.CREDS_FILE_PATH]):
            missing_class_attrs = ", ".join(
                [
                    attr
                    for attr in (
                        self.ACCESS_TOKEN_ENDPOINT,
                        self.BASE_URL,
                        self.CREDS_FILE_PATH,
                    )
                    if not attr
                ]
            )
            raise ValueError(
                "All class attributes need to be set in child class:"
                f" {missing_class_attrs}"
            )

        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        self.access_token_expiry_threshold = access_token_expiry_threshold
        self.log_requests = log_requests
        self.creds_cache_path = creds_cache_path or self.CREDS_FILE_PATH

        if logger:
            self.logger = logger
        else:
            self.logger = getLogger(__name__)
            self.logger.setLevel(DEBUG)
            add_stream_handler(self.logger)

        self._credentials = None

        self.temp_auth_server = None

    def _get(self, url, params=None):
        """Wrapper for GET requests which covers authentication, URL parsing, etc. etc.

        Args:
            url (str): the URL path to the endpoint (not necessarily including the
             base URL)
            params (dict): the parameters to be passed in the HTTP request

        Returns:
            Response: the response from the HTTP request
        """

        if url.startswith("/"):
            url = f"{self.BASE_URL}{url}"

        if self.log_requests:
            self.logger.debug(
                "GET %s with params %s", url, dumps(params or {}, default=str)
            )

        res = get(
            url,
            headers=self.request_headers,
            params=params or {},
        )

        res.raise_for_status()

        return res

    def _load_local_credentials(self):
        try:
            with open(self.creds_cache_path, encoding="UTF-8") as fin:
                self._credentials = load(fin).get(self.client_id, {})
        except FileNotFoundError:
            self.logger.info("Unable to find local creds file")
            self._credentials = {}

    def exchange_auth_code(self, code):
        """Allows first-time (or repeated) authentication

        Args:
            code (str): the authorization code returned from the auth flow
        """

        res = post(
            self.ACCESS_TOKEN_ENDPOINT,
            data={
                "grant_type": "authorization_code",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "redirect_uri": self.redirect_uri,
                "code": code,
            },
        )

        res.raise_for_status()

        self.credentials = res.json()

    def get_json_response(self, url, params=None):
        """Gets a simple JSON object from a URL

        Args:
            url (str): the API endpoint to GET
            params (dict): the parameters to be passed in the HTTP request

        Returns:
            dict: the JSON from the response
        """
        return self._get(url, params=params).json()

    def refresh_access_token(self):
        """Uses the cached refresh token to submit a request to TL's API for a new
        access token"""

        self.logger.info("Refreshing access token")

        res = post(
            self.ACCESS_TOKEN_ENDPOINT,
            data={
                "grant_type": "refresh_token",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "refresh_token": self._credentials.get("refresh_token"),
            },
        )

        res.raise_for_status()

        # Maintain any existing credential values in the dictionary whilst updating
        # new ones
        self.credentials = {
            **self._credentials,
            **res.json(),
        }

    @property
    def access_token(self):
        """
        Returns:
            str: the access token for this bank's API
        """
        return self.credentials.get("access_token")

    @property
    def access_token_has_expired(self):
        """Decodes the JWT access token and evaluates the expiry time

        Returns:
            bool: has the access token expired?
        """
        try:
            expiry_epoch = decode(
                # can't use self.access_token here, as that uses self.credentials,
                # which in turn (recursively) checks if the access token has expired
                self._credentials.get("access_token"),
                options={"verify_signature": False},
            ).get("exp", 0)

            return (expiry_epoch - self.access_token_expiry_threshold) < int(time())
        except DecodeError:
            # treat invalid token as expired, so we get a new one
            return True

    @property
    def credentials(self):
        """Attempts to retrieve credentials from local cache, creates new ones if
        they're not found.

        Returns:
            dict: the credentials for the chosen bank

        Raises:
            ValueError: if the state token returned from the request doesn't match the
             expected value
        """
        self._load_local_credentials()

        if not self._credentials:
            self.logger.info("Performing first time login")
            state_token = "".join(choice(ascii_letters) for _ in range(32))
            # pylint: disable=line-too-long
            auth_link = f"https://auth.monzo.com/?client_id={self.client_id}&redirect_uri={self.redirect_uri}&response_type=code&state={state_token}"  # noqa E501
            self.logger.debug("Opening %s", auth_link)
            open_browser(auth_link)

            self.temp_auth_server = TempAuthServer(__name__)

            request_args = self.temp_auth_server.wait_for_request(
                "/get_auth_code", kill_on_request=True
            )

            if state_token != request_args.get("state"):
                raise ValueError(
                    "State token received in request doesn't match expected value"
                )

            code = request_args.get("code")

            self.exchange_auth_code(code)

        if self.access_token_has_expired:
            self.refresh_access_token()

        return self._credentials

    @credentials.setter
    def credentials(self, value):
        self._credentials = value

        try:
            with open(
                force_mkdir(self.creds_cache_path, path_is_file=True), encoding="UTF-8"
            ) as fin:
                all_credentials = load(fin)
        except FileNotFoundError:
            all_credentials = {}

        all_credentials[self.client_id] = self._credentials

        with open(self.creds_cache_path, "w", encoding="UTF-8") as fout:
            dump(all_credentials, fout)

    @property
    def request_headers(self):
        """
        Returns:
            dict: auth headers for HTTP requests
        """
        return {"Authorization": f"Bearer {self.access_token}"}

    @property
    def refresh_token(self):
        """
        Returns:
            str: the TL API refresh token
        """
        return self.credentials.get("refresh_token")


class GoogleClient:
    """Custom client for interacting with the Google APIs

    Args:
        project (str): the name of the project which this client is being used for
        scopes (list): a list of scopes the client can be given
        client_id_json_path (str): the path to the `client_id.json` file downloaded
         from Google's API Console
        creds_cache_path (str): file path for where to cache credentials
        access_token_expiry_threshold (int): the threshold for when the access token is
         considered expired
        logger (RootLogger): a logger to use throughout the client functions
    """

    DEFAULT_PARAMS = {
        "pageSize": "50",
    }

    def __init__(
        self,
        project,
        scopes=None,
        client_id_json_path=None,
        creds_cache_path=None,
        access_token_expiry_threshold=60,
        logger=None,
    ):
        self.project = project
        self.scopes = scopes or []
        self.client_id_json_path = client_id_json_path
        self.creds_cache_path = creds_cache_path or user_data_dir(
            file_name="google_api_creds.json"
        )
        self.access_token_expiry_threshold = access_token_expiry_threshold

        if logger:
            self.logger = logger
        else:
            self.logger = getLogger(__name__)
            self.logger.setLevel(DEBUG)
            add_stream_handler(self.logger)

        if not scopes:
            self.logger.warning(
                "No scopes set for Google client. Functionality will be limited."
            )

        self._all_credentials_json = {}
        self._session = None

        self.temp_auth_server = None

    def _list_items(self, method, url, list_key, *, params=None):
        """Generic method for listing items on Google's API(s)

        Args:
            method (callable): the Google client session method to use
            url (str): the API endpoint to send a request to
            list_key (str): the key to use in extracting the data from the response
            params (dict): any extra params to be passed in the request

        Returns:
            list: a list of dicts, each representing an item from the API
        """
        params = (
            {**self.DEFAULT_PARAMS, **params}
            if params
            else deepcopy(self.DEFAULT_PARAMS)
        )
        self.logger.info(
            "Listing all items at endpoint `%s` with params %s", url, dumps(params)
        )

        res = method(url, params=params)

        res.raise_for_status()

        item_list = res.json().get(list_key, [])

        while next_token := res.json().get("nextPageToken"):
            res = method(
                url,
                params={**params, "pageToken": next_token},
            )
            res.raise_for_status()
            item_list.extend(res.json().get(list_key, []))
            self.logger.debug("Found %i items so far", len(item_list))

        return item_list

    def delete_creds_file(self):
        """Delete the local creds file"""
        try:
            remove(self.creds_cache_path)
        except FileNotFoundError:
            pass

    def get_items(self, url, list_key="items", *, params=None):
        """Wrapper method for getting a list of items

        See Also:
            self._list_items: main worker method for this functionality
        """

        return self._list_items(self.session.get, url, list_key, params=params)

    def refresh_access_token(self):
        """Uses the cached refresh token to submit a request to TL's API for a new
        access token"""

        self.logger.info("Refreshing access token")

        res = post(
            "https://accounts.google.com/o/oauth2/token",
            data={
                "grant_type": "refresh_token",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "refresh_token": self.refresh_token,
            },
        )

        res.raise_for_status()

        # Maintain any existing credential values in the dictionary whilst updating
        # new ones
        self.credentials = {
            **self._all_credentials_json[self.project],
            **res.json(),
        }

    @property
    def access_token_has_expired(self):
        """Decodes the JWT access token and evaluates the expiry time

        Returns:
            bool: has the access token expired?
        """

        if self.project not in self._all_credentials_json:
            return True

        try:
            expiry_epoch = decode(
                # can't use self.access_token here, as that uses self.credentials,
                # which in turn (recursively) checks if the access token has expired
                self._all_credentials_json[self.project].get("access_token"),
                options={"verify_signature": False},
            ).get("exp", 0)

            return (expiry_epoch - self.access_token_expiry_threshold) < int(time())
        except DecodeError:
            # treat invalid token as expired, so we get a new one
            return True

    @property
    def client_id(self):
        """
        Returns:
            str: the current client ID
        """
        return self._all_credentials_json.get(self.project, {}).get("client_id")

    @property
    def client_secret(self):
        """
        Returns:
            str: the current client secret
        """
        return self._all_credentials_json.get(self.project, {}).get("client_secret")

    @property
    def credentials(self):
        """Gets creds as necessary (including first time setup) and authenticates them

        Returns:
            Credentials: authorized credentials for use in creating a session

        Raises:
            EOFError: when no data is successfully returned for the auth code (usually
             when running the script automatically)
            ValueError: same as above, but if the EOFError isn't raised
        """

        try:
            with open(
                force_mkdir(self.creds_cache_path, path_is_file=True), encoding="UTF-8"
            ) as fin:
                self._all_credentials_json = load(fin)
        except FileNotFoundError:
            self.logger.info("Unable to find local creds file")
            self._all_credentials_json = {}

        if self.project not in self._all_credentials_json:
            self.logger.info(
                "Performing first time login for project `%s`", self.project
            )

            self.client_id_json_path = self.client_id_json_path or input(
                "Download your Client ID JSON from https://console.cloud.google.com/"
                f"apis/credentials?project={self.project} and paste the file path"
                " here: "
            )

            flow = Flow.from_client_secrets_file(
                self.client_id_json_path,
                scopes=self.scopes,
                redirect_uri="http://localhost:5001/get_auth_code",
            )

            auth_url, _ = flow.authorization_url(access_type="offline")
            self.logger.debug("Opening %s", auth_url)
            open_browser(auth_url)

            self.temp_auth_server = TempAuthServer(__name__)

            request_args = self.temp_auth_server.wait_for_request(
                "/get_auth_code", kill_on_request=True
            )

            flow.fetch_token(code=request_args.get("code"))

            self.credentials = {
                "token": flow.credentials.token,
                "refresh_token": flow.credentials.refresh_token,
                "id_token": flow.credentials.id_token,
                "scopes": flow.credentials.scopes,
                "token_uri": flow.credentials.token_uri,
                "client_id": flow.credentials.client_id,
                "client_secret": flow.credentials.client_secret,
            }

        if self.access_token_has_expired:
            self.refresh_access_token()

        return Credentials.from_authorized_user_info(
            self._all_credentials_json[self.project], self.scopes
        )

    @credentials.setter
    def credentials(self, value):
        """
        Args:
            value (dict): the new values to use for the creds for this project
        """
        self._all_credentials_json[self.project] = value

        with open(
            force_mkdir(self.creds_cache_path, path_is_file=True), "w", encoding="UTF-8"
        ) as fout:
            dump(self._all_credentials_json, fout)

    @property
    def refresh_token(self):
        """
        Returns:
            str: the current refresh token
        """
        return self._all_credentials_json.get(self.project, {}).get("refresh_token")

    @property
    def session(self):
        """Uses the Credentials object to sign in to an authorized Google API session

        Returns:
            AuthorizedSession: an active, authorized Google API session
        """
        if not self._session:
            self._session = AuthorizedSession(self.credentials)

        return self._session



"""Custom client for interacting with Google's Drive API"""

from datetime import datetime, timedelta, date
from enum import Enum
from json import dumps
from typing import List  # pylint: disable=unused-import

from pytz import timezone, UTC
from tzlocal import get_localzone

from wg_utilities.clients._generic import GoogleClient


class ResponseStatus(Enum):
    """Enumeration for event attendee response statuses"""

    ACCEPTED = "accepted"
    DECLINED = "declined"
    TENTATIVE = "tentative"
    UNCONFIRMED = "needsAction"
    UNKNOWN = "unknown"


class EventType(Enum):
    """Enumeration for event types"""

    DEFAULT = "default"
    FOCUS_TIME = "focusTime"
    OUT_OF_OFFICE = "outOfOffice"


class _GoogleCalendarEntity:
    """

    Args:
        json (dict): the JSON from Google's API
        google_client (GoogleCalendarClient): a Google client for use in other requests
    """

    DATE_FORMAT = "%Y-%m-%d"
    DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S%z"

    def __init__(self, json=None, *, google_client=None):
        self._json = json or {}
        self.google_client = google_client

    @property
    def id(self):
        """
        Returns:
            str: identifier of the entity
        """
        return self._json.get("id")

    @property
    def kind(self):
        """
        Returns:
            str: type of the resource (e.g. "calendar#calendar", "calendar#event")
        """
        return self._json.get("kind")

    @property
    def etag(self):
        """
        Returns:
            str: ETag of the resource
        """
        return self._json.get("etag")

    @property
    def summary(self):
        """
        Returns:
            str: title of the entity
        """
        return self._json.get("summary")

    @property
    def description(self):
        """
        Returns:
            str: description of the calendar. Optional
        """
        return self._json.get("description")

    @property
    def location(self):
        """
        Returns:
            str: geographic location of the calendar as free-form text. Optional
        """
        return self._json.get("location")


class Event(_GoogleCalendarEntity):
    """Class for Google Calendar events"""

    def __init__(self, json=None, *, calendar, google_client=None):
        super().__init__(json, google_client=google_client)
        self._json = json or {}
        self.google_client = google_client
        self.calendar = calendar

    def delete(self):
        """Deletes the event from the host calendar"""

        res = self.google_client.session.delete(
            f"{self.google_client.BASE_URL}/calendars/"
            f"{self.calendar.id}/events/{self.id}"
        )

        res.raise_for_status()

    @property
    def attachments(self):
        """
        Returns:
            list: file attachments for the event (max 25)
        """
        return self._json.get("attachments")

    @property
    def attendees(self):
        """
        Returns:
            list: the attendees of the event
        """
        return self._json.get("attendees", [])

    @property
    def attendees_omitted(self):
        """
        Returns:
            bool: whether attendees may have been omitted from the event's
             representation. When retrieving an event, this may be due to a restriction
             specified by the maxAttendee query parameter
        """
        return self._json.get("attendeesOmitted")

    @property
    def color_id(self):
        """
        Returns:
            str:the color of the event. This is an ID referring to an entry in the
             event section of the colors definition
        """
        return self._json.get("colorId")

    @property
    def conference_data(self):
        """
        Returns:
            dict: the conference-related information, such as details of a Google Meet
             conference
        """
        return self._json.get("conferenceData")

    @property
    def created(self):
        """
        Returns:
            datetime: creation time of the event (as a RFC3339 timestamp)
        """
        return self._json.get("created")

    @property
    def creator(self):
        """
        Returns:
            dict: the creator of the event
        """
        return self._json.get("creator")

    @property
    def description(self):
        """
        Returns:
            str: description of the event; can contain HTML
        """
        return self._json.get("description")

    @property
    def end(self):
        """
        Returns:
            dict: the (exclusive) end time of the event. For a recurring event, this
             is the end time of the first instance
        """
        return self._json.get("end")

    @property
    def end_datetime(self):
        """
        Returns:
            datetime: the datetime at which this event ends/ed
        """
        end_datetime = None

        if end_datetime_str := self._json.get("end", {}).get("dateTime"):
            end_datetime = datetime.strptime(
                end_datetime_str,
                self.DATETIME_FORMAT,
            )

        if end_date_str := self._json.get("end", {}).get("date"):
            end_datetime = datetime.strptime(
                end_date_str,
                self.DATE_FORMAT,
            )

        if end_datetime and end_datetime.tzinfo is None:
            end_datetime = end_datetime.replace(
                tzinfo=timezone(self._json.get("end", {}).get("timeZone", "UTC"))
            )

        return end_datetime

    @property
    def end_time_unspecified(self):
        """
        Returns:
            bool: whether the end time is actually unspecified
        """
        return self._json.get("endTimeUnspecified")

    @property
    def event_type(self):
        """
        Returns:
            EventType: specific type of the event
        """
        return EventType(self._json.get("eventType", "default"))

    @property
    def extended_properties(self):
        """
        Returns:
            dict: extended properties of the event
        """
        return self._json.get("extendedProperties")

    @property
    def guests_can_invite_others(self):
        """
        Returns:
            bool: whether attendees other than the organizer can invite others to the
             event
        """
        return self._json.get("guestsCanInviteOthers")

    @property
    def guests_can_modify(self):
        """
        Returns:
            bool: whether attendees other than the organizer can modify the event
        """
        return self._json.get("guestsCanModify")

    @property
    def guests_can_see_other_guests(self):
        """
        Returns:
            bool: whether attendees other than the organizer can see who the event's
             attendees are
        """
        return self._json.get("guestsCanSeeOtherGuests")

    @property
    def hangout_link(self):
        """
        Returns:
            str: an absolute link to the Google Hangout associated with this event
        """
        return self._json.get("hangoutLink")

    @property
    def html_link(self):
        """
        Returns:
            str: an absolute link to this event in the Google Calendar Web UI
        """
        return self._json.get("htmlLink")

    @property
    def ical_uid(self):
        """
        Returns:
            str: event unique identifier as defined in RFC5545. It is used to uniquely
             identify events across calendaring systems
        """
        return self._json.get("iCalUID")

    @property
    def locked(self):
        """
        Returns:
            bool: whether this is a locked event copy where no changes can be made to
             the main event fields "summary", "description", "location", "start", "end"
             or "recurrence"
        """
        return self._json.get("locked")

    @property
    def organizer(self):
        """
        Returns:
            dict: the organizer of the event
        """
        return self._json.get("organizer")

    @property
    def original_start_time(self):
        """
        Returns:
            dict: for an instance of a recurring event, this is the time at which this
             event would start according to the recurrence data in the recurring event
             identified by recurringEventId
        """
        return self._json.get("originalStartTime")

    @property
    def private_copy(self):
        """
        Returns:
            bool: if set to True, Event propagation is disabled
        """
        return self._json.get("privateCopy")

    @property
    def recurrence(self):
        """
        Returns:
            list: list of RRULE, EXRULE, RDATE and EXDATE lines for a recurring event,
             as specified in RFC5545
        """
        return self._json.get("recurrence")

    @property
    def recurring_event_id(self):
        """
        Returns:
            str: for an instance of a recurring event, this is the id of the recurring
             event to which this instance belongs
        """
        return self._json.get("recurringEventId")

    @property
    def reminders(self):
        """
        Returns:
            dict: information about the event's reminders for the authenticated user
        """
        return self._json.get("reminders")

    @property
    def response_status(self):
        """
        Returns:
            ResponseStatus: the response status for the authenticated user
        """
        for attendee in self.attendees:
            if attendee.get("self") is True:
                return ResponseStatus(attendee.get("responseStatus", "unknown"))

        # Own events don't always have attendees
        if self.creator.get("self") is True:
            return ResponseStatus.ACCEPTED

        return ResponseStatus("unknown")

    @property
    def sequence(self):
        """
        Returns:
            int: sequence number as per iCalendar
        """
        return self._json.get("sequence")

    @property
    def source(self):
        """
        Returns:
            dict: source from which the event was created
        """
        return self._json.get("source")

    @property
    def start(self):
        """
        Returns:
            dict: the (inclusive) start time of the event; for a recurring event, this
             is the start time of the first instance
        """
        return self._json.get("start")

    @property
    def start_datetime(self):
        """
        Returns:
            datetime: the datetime at which this event starts/ed
        """
        start_datetime = None

        if start_datetime_str := self._json.get("start", {}).get("dateTime"):
            start_datetime = datetime.strptime(
                start_datetime_str,
                self.DATETIME_FORMAT,
            )

        if start_date_str := self._json.get("start", {}).get("date"):
            start_datetime = datetime.strptime(
                start_date_str,
                self.DATE_FORMAT,
            )

        if start_datetime and start_datetime.tzinfo is None:
            start_datetime = start_datetime.replace(
                tzinfo=timezone(self._json.get("start", {}).get("timeZone", "UTC"))
            )

        return start_datetime

    @property
    def status(self):
        """
        Returns:
            str: status of the event
        """
        return self._json.get("status")

    @property
    def transparency(self):
        """
        Returns:
            str: whether the event blocks time on the calendar
        """
        return self._json.get("transparency")

    @property
    def updated(self):
        """
        Returns:
            datetime: last modification time of the event
        """
        if updated_str := self._json.get("updated"):
            return datetime.strptime(updated_str, self.DATETIME_FORMAT)

        return None

    @property
    def visibility(self):
        """
        Returns:
            str: visibility of the event
        """
        return self._json.get("visibility")

    def __lt__(self, other):
        if self.start_datetime == other.start_datetime:
            return self.summary.lower() < other.summary.lower()

        return self.start_datetime < other.start_datetime

    def __gt__(self, other):
        if self.start_datetime == other.start_datetime:
            return self.summary.lower() > other.summary.lower()

        return self.start_datetime > other.start_datetime

    def __str__(self):
        try:
            return (
                f"{self.summary} ("
                f"{self.start_datetime.strftime('%Y-%m-%d %H:%M:%S')} - "
                f"{self.end_datetime.strftime('%Y-%m-%d %H:%M:%S')})"
            )
        except AttributeError:
            return self.summary + dumps(self.start)


class Calendar(_GoogleCalendarEntity):
    """Class for Google calendar instances"""

    def get_events(
        self,
        page_size=2500,
        order_by="updated",
        from_datetime=None,
        to_datetime=None,
        combine_recurring_events=False,
    ):
        """Retrieve events from the calendar according to a set of criteria

        Args:
            page_size (int): the number of records to return on a single response page
            order_by (Literal["updated", "startTime"]): the order of the events
             returned within the result
            from_datetime (datetime): lower bound (exclusive) for an event's end time
             to filter by
            to_datetime (datetime): upper bound (exclusive) for an event's start time
             to filter by
            combine_recurring_events (bool): whether to expand recurring events into
             instances and only return single one-off events and instances of recurring
             events, but not the underlying recurring events themselves

        Returns:
            List[Event]: a list of Event instances

        Raises:
            ValueError: if the time parameters are invalid
        """
        params = {
            "maxResults": page_size,
            "orderBy": order_by,
            "singleEvents": str(not combine_recurring_events),
        }
        if from_datetime or to_datetime:
            from_datetime = from_datetime or datetime.utcnow() - timedelta(days=90)
            to_datetime = to_datetime or datetime.utcnow()

            if from_datetime >= to_datetime:
                raise ValueError(
                    "If timeMax is set, timeMin must be smaller than timeMax, and vice"
                    " versa"
                )

            if from_datetime.tzinfo is None:
                from_datetime = from_datetime.replace(tzinfo=UTC)

            if to_datetime.tzinfo is None:
                to_datetime = to_datetime.replace(tzinfo=UTC)

            params["timeMin"] = from_datetime.strftime(self.DATETIME_FORMAT)
            params["timeMax"] = to_datetime.strftime(self.DATETIME_FORMAT)

        return [
            Event(item, calendar=self, google_client=self.google_client)
            for item in self.google_client.get_items(
                f"{self.google_client.BASE_URL}/calendars/{self.id}/events",
                params=params,
            )
        ]

    @property
    def time_zone(self):
        """
        Returns:
            str: the time zone of the calendar, formatted as an IANA Time Zone
             Database name, e.g. "Europe/Zurich". Optional
        """
        return self._json.get("timeZone")

    @property
    def conference_properties(self):
        """
        Returns:
            dict: Conferencing properties for this calendar, for example what types
             of conferences are allowed
        """
        return self._json.get("conferenceProperties")

    def __str__(self):
        return self.summary


class GoogleCalendarClient(GoogleClient):
    """Custom client specifically for Google's Drive API

    Args:
        project (str): the name of the project which this client is being used for
        scopes (list): a list of scopes the client can be given
        client_id_json_path (str): the path to the `client_id.json` file downloaded
         from Google's API Console
        creds_cache_path (str): file path for where to cache credentials
        access_token_expiry_threshold (int): the threshold for when the access token is
         considered expired
        logger (RootLogger): a logger to use throughout the client functions
    """

    BASE_URL = "https://www.googleapis.com/calendar/v3"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._primary_calendar = None

    def create_event(
        self,
        summary,
        start_datetime,
        end_datetime,
        tz=None,
        calendar=None,
        extra_params=None,
    ):
        """Create an event

        Args:
            summary (str): the summary (title) of the event
            start_datetime (Union[datetime, date]): when the event starts
            end_datetime (Union[datetime, date]): when the event ends
            tz (str): the timezone which the event is in (IANA database name)
            calendar (Calendar): the calendar to add the event to
            extra_params (dict): any extra params to pass in the request

        Returns:
            Event: a new event instance, fresh out of the oven

        Raises:
            TypeError: if the start/end datetime params are not the correct type
        """

        calendar = calendar or self.primary_calendar
        tz = tz or str(get_localzone())

        start_params = {
            "timeZone": tz,
        }

        if isinstance(start_datetime, datetime):
            start_params["dateTime"] = start_datetime.strftime(
                _GoogleCalendarEntity.DATETIME_FORMAT
            )
        elif isinstance(start_datetime, date):
            start_params["date"] = start_datetime.strftime(
                _GoogleCalendarEntity.DATE_FORMAT
            )
        else:
            raise TypeError("`start_datetime` must be either a date or a datetime")

        end_params = {
            "timeZone": tz,
        }

        if isinstance(end_datetime, datetime):
            end_params["dateTime"] = end_datetime.strftime(
                _GoogleCalendarEntity.DATETIME_FORMAT
            )
        elif isinstance(end_datetime, date):
            end_params["date"] = end_datetime.strftime(
                _GoogleCalendarEntity.DATE_FORMAT
            )
        else:
            raise TypeError("`end_datetime` must be either a date or a datetime")

        res = self.session.post(
            f"{self.BASE_URL}/calendars/{calendar.id}/events",
            json={
                "summary": summary,
                "start": start_params,
                "end": end_params,
                **(extra_params or {}),
            },
        )

        res.raise_for_status()

        return Event(res.json(), calendar=calendar, google_client=self)

    def delete_event(self, event_id, calendar=None):
        """Deletes an event from a calendar

        Args:
            event_id (str): the ID of the event to delete
            calendar (Calendar): the calendar being updated
        """
        calendar = calendar or self.primary_calendar

        res = self.session.delete(
            f"{self.BASE_URL}/calendars/{calendar.id}/events/{event_id}"
        )

        res.raise_for_status()

    def get_event(self, event_id, calendar=None):
        """Get a specific event by ID

        Args:
            event_id (str): the ID of the event to delete
            calendar (Calendar): the calendar being updated

        Returns:
            Event: an Event instance with all relevant attributes
        """
        calendar = calendar or self.primary_calendar
        res = self.session.get(
            f"{self.BASE_URL}/calendars/{calendar.id}/events/{event_id}"
        )

        res.raise_for_status()

        return Event(res.json(), calendar=calendar, google_client=self)

    @property
    def calendar_list(self):
        """
        Returns:
            list: a list of Calendar instances that the user has access to
        """
        return [
            Calendar(cal_json, google_client=self)
            for cal_json in self.get_items(
                f"{self.BASE_URL}/users/me/calendarList",
            )
        ]

    @property
    def primary_calendar(self):
        """
        Returns:
            Calendar: the current user's primary calendar
        """
        if not self._primary_calendar:
            self._primary_calendar = Calendar(
                json=self.session.get(f"{self.BASE_URL}/calendars/primary").json(),
                google_client=self,
            )

        return self._primary_calendar



"""Useful functions"""

from .datetime_helpers import utcnow
from .file_management import user_data_dir, force_mkdir
from .processes import run_cmd
from .string_manipulation import cleanse_string
from .xml import get_nsmap
from .json import set_nested_value
from ._functions import chunk_list, flatten_dict, try_float



"""Custom client for interacting with Google's Fit API"""

from datetime import datetime

from wg_utilities.clients._generic import GoogleClient
from wg_utilities.functions.datetime_helpers import utcnow, DatetimeFixedUnit as DFUnit


class DataSource:
    """Class for interacting with Google Fit Data Sources

    Args:
        data_source_id (str): the unique ID of the data source
        google_client (GoogleClient): a GoogleClient instance, needed for getting
         DataSource info
    """

    DP_VALUE_KEY_LOOKUP = {"floatPoint": "fpVal", "integer": "intVal"}

    GOOGLE_CLIENT = None

    def __init__(self, data_source_id, *, google_client=None):
        self.data_source_id = data_source_id
        self.url = (
            f"{GoogleFitClient.BASE_URL}/users/me/dataSources/{self.data_source_id}"
        )

        self._description = None

        if isinstance(google_client, GoogleClient) and self.GOOGLE_CLIENT is None:
            # This is a class attribute, so it can be shared across all instances
            DataSource.GOOGLE_CLIENT = google_client

    @property
    def description(self):
        """
        Returns:
            dict: the JSON description of this data source
        """
        if not self._description:
            self._description = self.google_client.session.get(self.url).json()

        return self._description

    def sum_data_points_in_range(self, from_datetime=None, to_datetime=None):
        """Gets the sum of data points in the given range: if no `from_datetime` is
        provided, it defaults to the start of today; if no `to_datetime` is provided
        then it defaults to now.

        Args:
            from_datetime (datetime): lower boundary for step count
            to_datetime (datetime): upper boundary for step count

        Returns:
            list: a list of data point in the given range
        """

        from_nano = (
            int(from_datetime.timestamp() * 1000000000)
            if from_datetime
            else int(
                datetime.today()
                .replace(hour=0, minute=0, second=0, microsecond=0)
                .timestamp()
                / DFUnit.NANOSECOND.value
            )
        )

        to_nano = (
            int(to_datetime.timestamp() * 1000000000)
            if to_datetime
            else utcnow(DFUnit.NANOSECOND)
        )

        data_points = self.google_client.get_items(
            f"{self.url}/datasets/{from_nano}-{to_nano}",
            "point",
        )

        count = 0
        for point in data_points:
            if (
                int(point["startTimeNanos"]) >= from_nano
                and int(point["endTimeNanos"]) <= to_nano
            ):
                count += point["value"][0][self.data_point_value_key]

        return count

    @property
    def data_type_field_format(self):
        """
        Returns:
            str: the field format of this data source (i.e. "integer" or "floatPoint")
        """
        return self.description.get("dataType", {}).get("field", [{}])[0].get("format")

    @property
    def data_point_value_key(self):
        """
        Returns:
            str: the key to use when extracting data from a data point
        """
        return self.DP_VALUE_KEY_LOOKUP.get(self.data_type_field_format)

    @property
    def google_client(self):
        """

        Returns:
            GoogleClient: a GoogleClient instance, needed for getting DataSource info
        """
        return self.GOOGLE_CLIENT

    # pylint: disable=no-self-use
    @google_client.setter
    def google_client(self, value):
        """Sets the class attribute for the Google client to the given value

        Args:
            value (GoogleClient): a GoogleClient instance, needed for getting
             DataSource info
        """
        DataSource.GOOGLE_CLIENT = value


class GoogleFitClient(GoogleClient):
    """Custom client for interacting with the Google Fit API

    See Also:
        GoogleClient: the base Google client, used for authentication and common
         functions
    """

    BASE_URL = "https://www.googleapis.com/fitness/v1"

    def __init__(
        self,
        project,
        scopes=None,
        client_id_json_path=None,
        creds_cache_path=None,
        access_token_expiry_threshold=60,
        logger=None,
    ):
        super().__init__(
            project,
            scopes=scopes,
            client_id_json_path=client_id_json_path,
            creds_cache_path=creds_cache_path,
            access_token_expiry_threshold=access_token_expiry_threshold,
            logger=logger,
        )
        self.data_sources = {}

    def get_data_source(self, data_source_id):
        """Gets a data source based on its UID. DataSource instances are cached for the
         lifetime of the GoogleClient instance

        Args:
            data_source_id (str): the UID of the data source

        Returns:
            DataSource: an instance, ready to use!
        """

        if not (data_source := self.data_sources.get(data_source_id)):
            data_source = DataSource(data_source_id, google_client=self)
            self.data_sources[data_source_id] = data_source

        return data_source



"""Helper functions specifically for parsing/manipulating XML"""
from logging import getLogger, DEBUG
from lxml.etree import fromstring

from wg_utilities.loggers import add_stream_handler

LOGGER = getLogger(__name__)
LOGGER.setLevel(DEBUG)
add_stream_handler(LOGGER)


def get_nsmap(*, root=None, xml_doc=None, warn_on_defaults=False):
    """Get the namespace map for an XML document

    Args:
        root (Element): an lxml Element from an XML document
        xml_doc (str): a raw XML document
        warn_on_defaults (bool): log a warning when an empty prefix is found and
         converted to a default value

    Returns:
        dict: a namespace mapping for the provided XML

    Raises:
        ValueError: if neither argument is provided
    """
    if not (root or xml_doc):
        raise ValueError("One of `root` or `xml_doc` should be non-null")

    root = root or fromstring(xml_doc.encode())
    nsmap = {}

    default_count = 0
    processed_urls = set()
    for prefix, url in root.xpath("//namespace::*"):
        if url in processed_urls:
            continue

        if prefix is None:
            prefix = f"default_{default_count}"
            default_count += 1
            if warn_on_defaults:
                LOGGER.warning(
                    "Adding namespace url `%s` with prefix key `%s`", url, prefix
                )

        nsmap[prefix] = url
        processed_urls.add(url)

    return nsmap



"""Useful clients for commonly accessed APIs/services"""

from .google_calendar import GoogleCalendarClient
from .google_drive import GoogleDriveClient
from .google_fit import GoogleFitClient
from .google_photos import GooglePhotosClient
from .monzo import MonzoClient
from .spotify import SpotifyClient
from .truelayer import TrueLayerClient
